
#include "PenguinSoccer.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <string>

int Pos::manhattan_distance(const Pos &rhs) const
{
    return std::max(abs(rhs.x - x), abs(rhs.y - y));
}

Direction turn_towards(Direction d, Direction target)
{
    switch (d) {
        case NORTH:
            switch (target) {
                case NORTHWEST: case WEST: case SOUTHWEST: return NORTHWEST;
                default: return NORTHEAST;
            }
        case NORTHEAST:
            switch (target) {
                case NORTH: case NORTHWEST: case WEST: return NORTH;
                default: return EAST;
            }
        case EAST:
            switch (target) {
                case NORTHEAST: case NORTH: case NORTHWEST: return NORTHEAST;
                default: return SOUTHEAST;
            }
        case SOUTHEAST:
            switch (target) {
                case EAST: case NORTHEAST: case NORTH: return EAST;
                default: return SOUTH;
            }
        case SOUTH:
            switch (target) {
                case SOUTHEAST: case EAST: case NORTHEAST: return SOUTHEAST;
                default: return SOUTHWEST;
            }
        case SOUTHWEST:
            switch (target) {
                case SOUTH: case SOUTHEAST: case EAST: return SOUTH;
                default: return WEST;
            }
        case WEST:
            switch (target) {
                case SOUTHWEST: case SOUTH: case SOUTHEAST: return SOUTHWEST;
                default: return NORTHWEST;
            }
        case NORTHWEST:
            switch (target) {
                case WEST: case SOUTHWEST: case SOUTH: return WEST;
                default: return NORTH;
            }
        default:
            assert(false);
            return NORTH;
    }
}

Pos & Pos::operator += (Direction d)
{
    const int dx = ((int)d / 3) - 1;
    const int dy = ((int)d % 3) - 1;
    assert(-1 <= dx && dx <= 1);
    assert(-1 <= dy && dy <= 1);
    assert(dx != 0 || dy != 0);
    x += dx;
    y += dy;
    return *this;
}

bool Pos::off_board() const
{
    if (x < 0 || 8 <= x) return true;
    if (y < 0 || 8 <= y) return true;
    return false;
}


Board::Board()
{
    for (int w=0; w < 2; ++w) {
        for (int s=0; s < 3; ++s) {
            penguin[w][s].color = w;
            penguin[w][s].size = s;
            penguin[w][s].on_board = false;
        }
    }
    ball_on_board = false;
    attacker = 0;
}


bool Board::has_ball(const Penguin &p) const
{
    if (!ball_on_board) return false;
    if (!p.on_board) return false;
    if (p.pos == ball_pos) {
        assert(p.standing);
        return true;
    }
    return false;
}


Penguin *Board::penguinAt(const Pos &pos)
{
    for (int w=0; w < 2; ++w) {
        for (int s=0; s < 3; ++s) {
            Penguin *p = &penguin[w][s];
            if (p->on_board && p->pos == pos)
              return p;
        }
    }
    return NULL;
}

const Penguin *Board::penguinAt(const Pos &pos) const
{
    for (int w=0; w < 2; ++w) {
        for (int s=0; s < 3; ++s) {
            const Penguin *p = &penguin[w][s];
            if (p->on_board && p->pos == pos)
              return p;
        }
    }
    return NULL;
}


bool Board::game_over() const
{
    if (!ball_on_board) return false;
    return ball_pos.isGoal(0) || ball_pos.isGoal(1);
}

int findattacker(const Board &b)
{
    return b.attacker;
}


bool is_valid_move(const Board &b, const Move &m)
{
    Board nb = b;
    return apply_move(nb, m);
}

static bool topple_penguin(Board &b, Penguin &p, Direction dir);
static const char *piece2str(int,int);

bool apply_move(Board &b, const Move &m)
{
    const int who = b.attacker;
    b.attacker = 1-who;
    assert(0 <= who && who <= 1);
    assert(0 <= m.penguin_id && m.penguin_id <= 2);
    Penguin &p = b.penguin[who][m.penguin_id];
    if (m.kind == STAND) {
        if (!p.on_board) return false;
        if (p.standing) return false;
        p.standing = true;
        assert(!b.has_ball(p));
        return true;
    } else if (m.kind == KICK) {
        if (!b.has_ball(p)) return false;
        assert(p.on_board && p.standing);
        const int distance_kicked = p.size + 1;
        for (int i=0; i < distance_kicked; ++i) {
            Pos nbpos = b.ball_pos;
            nbpos += m.dir;
            /* The ball doesn't leave the board. */
            if (nbpos.off_board()) break;
            /* Okay, the ball will enter this square. */
            b.ball_pos = nbpos;
            /* The ball may be caught by a penguin. */
            if (Penguin *catcher = b.penguinAt(nbpos)) {
                catcher->standing = true;
                break;
            }
        }
        /* Kicking the ball against a wall isn't allowed
         * because it would lead to weirdness. */
        if (b.ball_pos == p.pos)
          return false;
        /* The kicking penguin falls over. */
        p.standing = false;
        p.dir = m.dir;
        return true;
    } else {
        assert(m.kind == SLIDE);
        /* A penguin with the ball may not slide. */
        if (b.has_ball(p)) return false;
        Pos nppos;
        if (!p.on_board) {
            switch (p.color) {
                case 0: /* Black penguins start at (0,0). */
                    nppos.x = -1; nppos.y = -1;
                    if (m.dir == NORTHEAST) nppos.y = 0;
                    else if (m.dir == NORTHWEST) nppos.x = 0;
                    else if (m.dir == NORTH) { /* nothing */ }
                    else return false;
                    break;
                case 1:
                    nppos.x = 8; nppos.y = 8;
                    if (m.dir == SOUTHEAST) nppos.x = 7;
                    else if (m.dir == SOUTHWEST) nppos.y = 7;
                    else if (m.dir == SOUTH) { /* nothing */ }
                    else return false;
                    break;
            }
            p.pos = nppos;
            p.on_board = true;
        } else {
            if (!p.standing && m.dir != p.dir) return false;
            nppos = p.pos;
        }
        p.standing = false;
        p.dir = m.dir;
        const int distance_slid = 3 - p.size;
        for (int i=0; i < distance_slid; ++i) {
            nppos += m.dir;
            /* The penguin doesn't leave the board. */
            if (nppos.off_board()) break;
            /* You aren't allowed to enter the opposing home square. */
            if (nppos.isHome(1-who)) return false;
            /* You aren't allowed to tackle anyone who doesn't have
             * the ball. */
            Penguin *tacklee = b.penguinAt(nppos);
            if (tacklee != NULL && tacklee->color == who) return false;
            if (tacklee != NULL && !b.has_ball(*tacklee)) return false;
            /* Okay, the penguin will enter this square. */
            if (!b.ball_on_board && nppos.isCenter()) {
                /* The ball is collected by the first penguin to reach
                 * any of the central four squares. */
                b.ball_on_board = true;
                assert(!b.penguinAt(nppos));
                p.pos = nppos;
                p.standing = true;
                b.ball_pos = nppos;
                assert(b.has_ball(p));
                return true;
            } else if (b.ball_on_board && nppos == b.ball_pos) {
                /* A penguin who reaches the ball will stop and stand up. */
                if (tacklee != NULL) {
                    if (!topple_penguin(b, *tacklee, m.dir))
                      return false;
                }
                p.pos = nppos;
                p.standing = true;
                assert(b.has_ball(p));
                return true;
            }
            if (p.dir != m.final_dir) {
                p.dir = turn_towards(p.dir, m.final_dir);
            }
            p.pos = nppos;
        }
        /* If m.final_dir was inconsistent, report an error.
         * We should have already handled the cases where
         * the sliding penguin stood back up. */
        assert(!p.standing);
        if (p.dir != m.final_dir)
          return false;
        /* Otherwise, we succeeded. */
        return true;
    }
}

static bool topple_penguin(Board &b, Penguin &p, Direction dir)
{
    /* The tackled penguin is pushed one square
     * in the direction the tackler was moving,
     * and oriented in the direction of the push. */
    Pos nppos = p.pos;
    nppos += dir;
    if (nppos.off_board()) {
        p.on_board = false;
    } else if (nppos.isHome(1-p.color)) {
        /* A penguin may never occupy an opposing home cell,
         * even if it is pushed there. */
        return false;
    } else {
        Penguin *p2 = b.penguinAt(nppos);
        if (p2 != NULL)
          topple_penguin(b, *p2, dir);
        p.pos = nppos;
        p.standing = false;
        p.dir = dir;
    }
    return true;
}


static const char *piece2str(int color, int size)
{
    assert(color == 0 || color == 1);
    assert(0 <= size && size <= 2);
    static char buf[3];
    buf[0] = "bw"[color];
    buf[1] = "123"[size];
    buf[2] = '\0';
    return buf;
}

static const char *dir2str(Direction d)
{
    switch (d) {
        case NORTHWEST: return "NW";
        case NORTH: return "N";
        case NORTHEAST: return "NE";
        case EAST: return "E";
        case SOUTHEAST: return "SE";
        case SOUTH: return "S";
        case SOUTHWEST: return "SW";
        case WEST: return "W";
        default: assert(false); return "none";
    }
}


void Board::display() const
{
    for (int i=0; i < 8; ++i) {
        char window[4];
        window[0] = ' ';
        for (int j=0; j < 8; ++j) {
            strcpy(window+1, ".. ");
            Pos pos(j,7-i);
            const Penguin *p = penguinAt(pos);
            const bool has_ball = ball_on_board && (ball_pos == pos);
            const bool is_goal = pos.isGoal(0) || pos.isGoal(1);
            if (is_goal) {
                if (window[0] != ')') window[0] = '[';
                window[3] = ']';
            }
            if (p != NULL) {
                if (has_ball) {
                    window[0] = '(';
                    window[3] = ')';
                }
                window[1] = "bw"[p->color];
                window[2] = "123"[p->size];
            } else {
                if (has_ball) {
                    window[1] = '(';
                    window[2] = ')';
                }
            }
            putchar(window[0]);
            putchar(window[1]);
            putchar(window[2]);
            window[0] = window[3];
        }
        putchar(window[3]);

        if (1 <= i && i <= 6) {
            const Penguin &p = penguin[(i-1)/3][(i-1)%3];
            printf("    %s:", piece2str(p.color, p.size));
            if (!p.on_board) {
                printf(" off board");
            } else if (p.standing) {
                printf(" standing");
                if (has_ball(p))
                  printf(" (with ball)");
            } else {
                printf(" %s", dir2str(p.dir));
            }
        }
        putchar('\n');
    }
}

std::string Board::toComparableString() const
{
    std::string result;
    for (int w=0; w < 2; ++w) {
        for (int s=0; s < 3; ++s) {
            const Penguin &p = penguin[w][s];
            if (!p.on_board) {
                result += "O";
            } else {
                result += ('a' + p.pos.x);
                result += ('a' + p.pos.y);
                if (p.standing)
                  result += 'S';
                else
                  result += ('i' + (int)p.dir);
            }
        }
    }
    if (ball_on_board) {
        result += ('A' + ball_pos.x);
        result += ('A' + ball_pos.y);
    }
    return result;
}

std::string Move::toString(int who) const
{
    std::string result = piece2str(who, penguin_id);
    if (kind == STAND) {
        result += " stand";
    } else if (kind == KICK) {
        result += " kick ";
        result += dir2str(dir);
    } else if (kind == SLIDE) {
        result += " slide ";
        result += dir2str(dir);
        if (final_dir != dir) {
            result += " finishing ";
            result += dir2str(final_dir);
        }
    }
    return result;
}

static bool scan_direction(const char *&text, Direction &dir)
{
    if (!strncmp(text, "nw", 2) || !strncmp(text, "NW", 2)) {
        text += 2; dir = NORTHWEST; return true;
    } else if (!strncmp(text, "ne", 2) || !strncmp(text, "NE", 2)) {
        text += 2; dir = NORTHEAST; return true;
    } else if (!strncmp(text, "sw", 2) || !strncmp(text, "SW", 2)) {
        text += 2; dir = SOUTHWEST; return true;
    } else if (!strncmp(text, "se", 2) || !strncmp(text, "SE", 2)) {
        text += 2; dir = SOUTHEAST; return true;
    } else if (*text == 'n' || *text == 'N') {
        text += 1; dir = NORTH; return true;
    } else if (*text == 's' || *text == 'S') {
        text += 1; dir = SOUTH; return true;
    } else if (*text == 'e' || *text == 'E') {
        text += 1; dir = EAST; return true;
    } else if (*text == 'w' || *text == 'W') {
        text += 1; dir = WEST; return true;
    } else {
        return false;
    }
}

/* We scan moves with the following syntax.
 * It's intentionally terser than English, but
 * also less redundant than the SDG notation.
 *   2!     -- penguin 2 (papa) stands up
 *   1sw    -- penguin 1 (baby) slides southwest
 *   1sw se -- penguin 1 slides SW; finishes facing SE
 * But if penguin 1 is holding the ball, then
 *   1sw    -- penguin 1 (baby) kicks southwest
 */
bool Move::scan(const char *text, const Board &b)
{
    if (*text < '1' || '3' < *text) return false;
    penguin_id = (*text++ - '1');
    if (!strcmp(text, "!")) {
        kind = STAND;
        return true;
    } else {
        if (!scan_direction(text, dir)) return false;
        if (b.has_ball(b.penguin[b.attacker][penguin_id])) {
            kind = KICK;
            return (*text == '\0');
        } else {
            kind = SLIDE;
            final_dir = dir;
            if (*text == '\0') return true;
            while (*text == ' ') ++text;
            if (!scan_direction(text, final_dir)) return false;
            return (*text == '\0');
        }
    }
}

